from pathlib import Path
from typing import List

from ingest_envoy.manifest_components import OutputScienceProduct, AncillaryProduct, WEBLOG_FILENAME
from ingest_envoy.utilities import AncillaryProductType


class ImageIngestionProductsFinder:
    """Finds ancillary science products and other ancillary products needed for image ingestion"""

    def __init__(self, staging_source_dir: Path):
        self.staging_source_dir = staging_source_dir
        self.files_found = [file for file in self.staging_source_dir.iterdir()]
        self.science_products = self._find_output_science_products()
        self.ancillary_products = self._find_other_ancillary_products()

    def _find_output_science_products(self) -> List[OutputScienceProduct]:
        """
        Find the ancillary products belonging to the science product*
        in the staging dir.

        * in our example there is exactly one science product with a couple of ancillary
        products, so for now the dict we return will contain only one science product.
        if this changes for some new type of science product type, we'll have to figure
        out how to determine which ancillary product(s) belong to which science product.


        :return:
        """
        sp_image_file = [
            file
            for file in self.files_found
            if file.name.endswith(".fits") and "rms" not in file.name
        ][0]
        image_files = [
            file for file in self.files_found if "rms" in file.name or file.name.endswith(".png")
        ]

        sp_aps = []
        for image_file in image_files:
            sp_ap = self._build_ancillary_image_science_product(image_file)
            sp_aps.append(sp_ap)

        science_product = OutputScienceProduct(
            type=AncillaryProductType.QUICKLOOK_IMAGE,
            filename=str(sp_image_file),
            ancillary_products=sp_aps,
        )

        return [science_product]

    def _find_other_ancillary_products(self) -> List[AncillaryProduct]:
        """
        TODO
        Find the "other" ancillary image products in the staging dir: there should be a weblog
        and a pipeline artifacts tar. (The ingestion artifacts tar will be produced during the
        building of the manifest.)

        :return:
        """

        ancillary_products = []
        try:
            weblog = [file for file in self.files_found if file.name == WEBLOG_FILENAME][0]
            ancillary_products.append(
                AncillaryProduct(
                    type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(weblog)
                )
            )
        except Exception as exc:
            # TODO which exception will this be?
            raise FileNotFoundError(f"No weblog found in {self.staging_source_dir}") from exc

        try:
            pipeline_artifacts_tar = [
                file
                for file in self.files_found
                if file.name.endswith("tar") and "ingestion_artifacts" not in file.name
            ][0]
            ancillary_products.append(
                AncillaryProduct(
                    type=AncillaryProductType.PIPELINE_ARTIFACTS,
                    filename=str(pipeline_artifacts_tar),
                )
            )
        except Exception as exc:
            # TODO which exception will this be?
            raise FileNotFoundError(
                f"No pipeline artifacts found in {self.staging_source_dir}"
            ) from exc

        return ancillary_products

    def _build_ancillary_image_science_product(self, file: Path):
        """
        Image science products will have ancillary products of their very own,
        distinct from other ancillary products that might be in the staging dir.

        :param file: a possible ancillary image product
        :return: the corresponding AncillaryProduct, if applicable
        """
        filename = str(file)
        if "image" in file.name:
            if file.name.endswith(".png"):
                return AncillaryProduct(type=AncillaryProductType.THUMBNAIL_IMG, filename=filename)
            elif file.name.endswith(".fits"):
                if "rms" in file.name:
                    return AncillaryProduct(
                        type=AncillaryProductType.QUICKLOOK_RMS_IMAGE, filename=filename
                    )
                else:
                    return AncillaryProduct(
                        type=AncillaryProductType.QUICKLOOK_IMAGE, filename=filename
                    )
