""" Tests for building image ingestion manifest """

from pathlib import Path
from typing import List

# pylint: disable=C0103, E0401, E0402, R1721

import json
import pytest

from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import Telescope, AncillaryProductType, ScienceProductType

# NOT unused. IJ is dumb. KEEP THIS.
from .conftest import ingest_path

from ingest_envoy.manifest_components import (
    WEBLOG_FILENAME,
    ManifestParameters,
    ParamsKey,
    OutputGroup,
    OutputScienceProduct,
    AncillaryProduct,
    InputGroup,
    InputScienceProduct,
)

IMG_MANIFEST_FILENAMES = [
    "image_metadata_2021_05_21_T10_17_19.180.json",
    "VLASS1.1.ql.T01t01.J000228-363000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits",
    "VLASS1.1.ql.T01t01.J000228-363000.10.2048.v1.I.iter1.image.pbcor.tt0.rms.subim.fits",
    WEBLOG_FILENAME,
    "ingestion_artifacts_2019_07_30_T13_03_00.936.tar",
]
OTHER_FILENAMES = [
    "unknown.pipeline_manifest.xml",
    "unknown.aux_products.tgz",
    "casa_commands.log",
    "casa_pipescript.py",
    "totally_random_tar.tar",
]


def test_parameters_json_well_formed():
    """
    Make sure we get expected ManifestParameters for an image.

    :return:
    """
    params = manifest_parameters()

    params_json = params.to_json()
    # Only difference from EVLA CAL manifest is additional_metadata
    assert params_json[ParamsKey.ADDITIONAL_METADATA.value] == str(params.additional_metadata)

    # gotta be able to dump it; test will fail if dump runs into trouble
    json.dumps(params_json)


def test_output_science_prods_built_properly():
    """
    Output science products for an image ingestion contain ancillary products.
    Let's make sure we can build one of these OSPs.
    :return:
    """
    ap = rms_ancillary_prod()
    osp = osp_ql_with_ancillary(ap)

    assert osp.type == AncillaryProductType.QUICKLOOK_IMAGE
    assert (
        osp.filename
        == "VLASS1.1.ql.T01t01.J000228-363000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits"
    )
    assert len(osp.ancillary_products) == 1

    osp_ap = osp.ancillary_products[0]
    assert osp_ap == ap
    ap_json = ap.to_json()
    # if we can't dump it, we know it's no good
    ap_str = json.dumps(ap_json)

    osp_json = osp.to_json()
    assert osp_json["type"] == osp.type.value
    assert osp_json["filename"] == osp.filename

    op_str = json.dumps(osp_json)
    assert ap_str in op_str


def test_output_group_json_well_formed():
    """
    In addition to the ancillary_products section of the image ingestion manifest,
    output science products may themselves contain ancillary products.
    :return:
    """

    output_group = build_output_group()
    expected_og_json = output_group.to_json()

    sps = output_group.science_products

    # since these are image science products, we expect there to be
    # ancillary image products
    for sp in sps:
        assert len(sp.ancillary_products) > 0
    expected_sp_json = [sp.to_json() for sp in sps]

    other_aps = separate_aps()
    expected_aps_json = [ap.to_json() for ap in other_aps]

    og_dump = json.dumps(expected_og_json)
    aps_dump = json.dumps(expected_aps_json)
    sps_dump = json.dumps(expected_sp_json)
    assert sps_dump in og_dump
    assert aps_dump in og_dump
    assert aps_dump not in sps_dump


@pytest.mark.skip("TODO: won't work until output group creation is fixed")
def test_creates_expected_manifest(ingest_path: Path):
    """
    Did the image ingestion manifest builder make the manifest we expected?

    :return:
    """

    # fill the ingestion path with fake files
    populate_fake_ingest_path(ingest_path)

    locator = "uid://evla/calibration/3dfa528b-9870-46c9-a200-131dbac701cc"
    addl_md = AbstractTextFile(filename="image_metadata_2021_05_21_T10_17_19.180.json", content="")
    builder = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.IMAGE,
        locator=locator,
        telescope=Telescope.EVLA,
        additional_metadata=addl_md,
    )
    manifest, _ = builder.build()

    expected_params = manifest_parameters()
    expected_params.staging_source_dir = ingest_path
    assert manifest.parameters == expected_params
    assert manifest.input_group == InputGroup(science_products=[InputScienceProduct(locator)])
    sp_ap1 = AncillaryProduct(
        AncillaryProductType.QUICKLOOK_RMS_IMAGE,
        filename="VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.I.iter1.image.pbcor.tt0.rms.subim.fits",
    )
    sp_ap2 = AncillaryProduct(
        type=AncillaryProductType.THUMBNAIL_IMG,
        filename="VLASS2.1.ql.T08t09.J055438_113000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.png",
    )

    # make a quicklook image science product with a quicklook_rms and a thumbnail
    sci_prod = OutputScienceProduct(
        type=AncillaryProductType.QUICKLOOK_IMAGE,
        filename="VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits",
        ancillary_products=[sp_ap1, sp_ap2],
    )

    weblog_path = ingest_path / WEBLOG_FILENAME
    ap1 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(weblog_path)
    )
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS,
        filename="VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1.tar",
    )
    ap3 = AncillaryProduct(
        type=AncillaryProductType.INGESTION_ARTIFACTS,
        filename="ingestion_artifacts_2021_05_21_T10_17_19.275.tar",
    )

    expected_output_group = OutputGroup(
        science_products=[sci_prod], ancillary_products=[ap1, ap2, ap3]
    )

    actual_output_group = manifest.output_group
    expected_sci_prods = expected_output_group.science_products
    actual_sci_prods = actual_output_group.science_products
    assert len(actual_sci_prods) == len(expected_sci_prods) == 1

    expected_anc_prods = expected_sci_prods[0].ancillary_products
    actual_anc_prods = actual_sci_prods[0].ancillary_products
    assert len(actual_anc_prods) == len(expected_anc_prods)

    assert actual_output_group.science_products == expected_output_group.science_products
    assert actual_output_group.ancillary_products == expected_output_group.ancillary_products
    assert actual_output_group == expected_output_group


@pytest.mark.skip("TODO: won't work until output group creation is fixed")
def test_filters_files_as_expected(ingest_path: Path):
    """
    The image ingestion manifest should contain no references to additional files in a directory
    that aren't needed for the manifest.

    :param ingest_path: the staging directory
    :return:
    """
    # fill the ingestion path with fake files
    populate_fake_ingest_path(ingest_path)

    locator = "uid://evla/calibration/3dfa528b-9870-46c9-a200-131dbac701cc"
    addl_md = AbstractTextFile(filename="image_metadata_2021_05_21_T10_17_19.180.json", content="")
    builder = IngestionManifestBuilder(
        staging_source_dir=ingest_path,
        sp_type=ScienceProductType.IMAGE,
        locator=locator,
        telescope=Telescope.EVLA,
        additional_metadata=addl_md,
    )
    manifest, _ = builder.build()

    files_in_manifest = [manifest.parameters.additional_metadata]
    output_group = manifest.output_group
    for sp in output_group.science_products:
        files_in_manifest.append(sp.filename)
        for ap in sp.ancillary_products:
            files_in_manifest.append(ap.filename)
        for ap in output_group.ancillary_products:
            files_in_manifest.append(ap.filename)

    for filename in IMG_MANIFEST_FILENAMES:
        assert filename in files_in_manifest
    for filename in OTHER_FILENAMES:
        assert filename not in files_in_manifest


@pytest.mark.skip("TODO")
def test_writes_expected_output_files(ingest_path: Path):
    """
    Did the image ingestion manifest builder produce the output file(s) we expect?

    :param ingest_path:
    :return:
    """


# -----------------------------
#  U  T  I  L  I  T  I  E  S
# -----------------------------


def manifest_parameters() -> ManifestParameters:
    """
    Build a ManifestParameters for our tests

    :return: the manifest parameters we're expecting
    """
    return ManifestParameters(
        telescope=Telescope.EVLA,
        reingest=False,
        ngas_ingest=False,
        calibrate=False,
        staging_source_dir=Path(
            "/lustre/aoc/cluster/pipeline/dsoc-prod/stage_products/"
            + "VLASS2.1.ql.T08t09.J055438-113000.10.2048.v1"
        ),
        additional_metadata=AbstractTextFile(
            filename="image_metadata_2021_05_21_T10_17_19.180.json", content=""
        ),
    )


def rms_ancillary_prod() -> AncillaryProduct:
    """
    Build an AncillaryProduct for our tests

    :return: the product we're expecting
    """
    return AncillaryProduct(
        type=AncillaryProductType.QUICKLOOK_RMS_IMAGE,
        filename="VLASS1.1.ql.T01t01.J000228-363000.10.2048.v1.I.iter1.image"
        + ".pbcor.tt0.rms.subim.fits",
    )


def osp_ql_with_ancillary(ap: AncillaryProduct) -> OutputScienceProduct:
    """
    Build an OutputScienceProduct for our tests

    :return: the product we're expecting
    """
    return OutputScienceProduct(
        type=AncillaryProductType(AncillaryProductType.QUICKLOOK_IMAGE),
        filename="VLASS1.1.ql.T01t01.J000228-363000.10.2048.v1.I.iter1.image.pbcor.tt0.subim.fits",
        ancillary_products=[ap],
    )


def separate_aps() -> List[AncillaryProduct]:
    """
    Build a list of AncillaryProducts for our tests

    :return: the products we're expecting
    """
    ap1 = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=WEBLOG_FILENAME)
    ap2 = AncillaryProduct(
        type=AncillaryProductType.PIPELINE_ARTIFACTS,
        filename="uid____EVLA_ingestion_artifacts_b1ab328d-200e-4da4-86bf-514773f31e2b.tar",
    )
    ap3 = AncillaryProduct(
        type=AncillaryProductType.INGESTION_ARTIFACTS,
        filename="ingestion_artifacts_2019_07_30_T13_03_00.936.tar",
    )
    return [ap1, ap2, ap3]


def build_output_group() -> OutputGroup:
    """
    Build an OutputGroup for our tests

    :return: the output group we're expecting
    """
    ap = rms_ancillary_prod()
    osp = osp_ql_with_ancillary(ap)

    other_aps = separate_aps()
    ap_list = other_aps

    return OutputGroup(science_products=[osp], ancillary_products=ap_list)


def populate_fake_ingest_path(staging_dir: Path) -> List[Path]:
    """
    Create a directory containing fake image products, plus other stuff
    that we -don't- want to ingest.

    :param staging_dir: our temporary dir
    :return:
    """
    for filename in IMG_MANIFEST_FILENAMES:
        # ingestion artifacts tar is produced during manifest creation
        if not filename.startswith("ingestion_artifacts"):
            file = staging_dir / filename
            file.touch()
    for filename in OTHER_FILENAMES:
        file = staging_dir / filename
        file.touch()
    return [file for file in staging_dir.iterdir()]
